package Distros;
use strict;

require Exporter;
@Distros::ISA = qw(Exporter);
# Exported symbols
push @Distros::EXPORT, qw(
    &guess_platform &guess_package_manager &does_shell_know &trim &is_elf_arch_correct
    %known_commands %known_wrong_commands
);

#----------------------------------------------------------------------------

# Helps to determine platform name.
sub guess_platform($ )
{
    my $name = $_[0];

    if ( $name =~ /^(amd64|x86_64)$/)
        { return "x86-64" }
    elsif ( $name =~ /^(ia32|i[3-6]86|x86)$/ )
        { return "IA32" }
    elsif ( $name =~ /^(ia64)$/ )
        { return "IA64" }
    elsif ( $name =~ /^(ppc|ppc32)$/ )
        { return "PPC32" }
    elsif ( $name =~ /^(ppc64)$/ )
        { return "PPC64" }
    elsif ( $name =~ /^(s390)$/ )
        { return "S390" }
    elsif ( $name =~ /^(s390x)$/ )
        { return "S390X" }
    else
        { return "unknown"; }
}

#----------------------------------------------------------------------------

# Determines which package manager to use.
sub guess_package_manager {
    my $host_os = "";
    if( does_shell_know ("lsb_release") ) {
        $host_os = `lsb_release -s -d`;
    }

    $host_os =~ s/\"(.*)\"/$1/;
	chomp($host_os);

    if ( $host_os =~ /(suse|red\W?hat|asianux|mandriva)/i )
    {
        # Distros known to be rpm-based
        return "rpm";
    }
    elsif ( $host_os =~ /(ubuntu|debian)/i )
    {
        # Distros known to use apt and dpkg
		# (note that on some Debian-based systems, like Xandros, `lsb_release` gives us 'Debian')
        return "dpkg";
    }
	elsif ( does_shell_know ("emerge") ) {
		# Gentoo
		return "emerge";
	}
    else {
		# Do we have 'dpkg'? Some systems may have both rpm and dpkg,
		# and 'dpkg' is usually the default manager in this case
        if ( does_shell_know ("dpkg") ) {
            return "dpkg";
        }
        elsif ( does_shell_know ("rpm") ) {
            return "rpm";
        }
    }

    return "unknown";
}

#----------------------------------------------------------------------------

sub is_elf_arch_correct($$$ ) {
	my($arch, $elf_arch, $elf_arch_bit) = ($_[0],$_[1],$_[2]);

	if( $arch =~ /x86-64/i ) {
		if( $elf_arch !~ /86-64/ ) {
			return 0;
		}
	}
	elsif( $arch =~ /IA64/i ) {
		if( $elf_arch !~ /IA-64/ ) {
			return 0;
		}
	}
	elsif( $arch =~ /S390X/i ) {
		if( ($elf_arch !~ /390/) or ($elf_arch_bit ne "64") ) {
			return 0;
		}
	}
	elsif( $arch =~ /PPC64/i ) {
		if( $elf_arch !~ /PowerPC 64-bit/ and $elf_arch !~ /PowerPC64/ and ($elf_arch !~ /PowerPC/ or $elf_arch_bit ne "64") ) {
			return 0;
		}
	}
	elsif( $arch =~ /PPC32/i ) {
		if( ($elf_arch !~ /PowerPC/) or ($elf_arch =~ /PowerPC 64-bit/) or ($elf_arch =~ /PowerPC64/) ) {
			return 0;
		}
	}
	elsif( $arch =~ /IA32/i ) {
		if( $elf_arch !~ /Intel 80386/ ) {
			return 0;
		}
	}
	elsif( $arch =~ /S390/i ) {
		if( ($elf_arch !~ /390/) or ($elf_arch_bit ne "32") ) {
			return 0;
		}
	}

	return 1;
}

#----------------------------------------------------------------------------

# Checks whether the specified tool is known to the shell.
sub does_shell_know
{
    my $tool_name = shift;
    return (system ("which $tool_name >/dev/null 2>/dev/null") == 0);
}

# Trim string spaces.
sub trim($)
{
	my $string = shift;
	$string =~ s/^\s+//;
	$string =~ s/\s+$//;
	return $string;
}

#----------------------------------------------------------------------------

# Commands sometimes reported by 'file' to be 'shared objects'
our %known_commands = (
    "su" => 1,
    "at" => 1,
    "crontab" => 1,
    "sudo" => 1,
    "sendmail.sendmail" => 1,
    "sendmail" => 1
);

# Files sometimes reported by 'file' as 'scripts' but which are not scripts, indeed
our %known_wrong_commands = (
    "CVS/Root" => 1,
    "CVS/Entries" => 1
);


1;