#!/usr/bin/perl -w

########################################################################
# This script makes list of library usage in different Distributions
#  and also in LSB versions as pseudo-distributions.
#
# Author: Smachev Andrey <biga@ispras.ru>
########################################################################

use DBI;

use Env qw(LSBUSER LSBDBPASSWD LSBDB LSBDBHOST);

my @all_lsb = ("3.0", "3.1", "3.2", "4.0");
my %all_archs = (); # "IA32", "IA64", "PPC32", "PPC64", "S390", "X86-64", "S390X"

sub usage()
{
	print STDERR "Usage: $0 [outfile] \n";
	print STDERR "Environment variables should be set for connecting to the LSB DB:\n"
		."    LSBDB, LSBDBHOST, LSBUSER, LSBDBPASSWD.\n";
	die;
}

my $LibDistrs_file = ( $ARGV[0] or "LibDistrs" );
# TODO: $Distributions_file

my $trace=1; # trace SQL statments

if ( !defined $LSBDBHOST ) { $LSBDBHOST = "localhost"; }
if ( !defined $LSBDB ) { $LSBDB = "lsb"; }
if ( !defined $LSBUSER ) { $LSBUSER = "lsbuser"; }
if ( !defined $LSBDBPASSWD ) { $LSBDBPASSWD = ""; }

# Establish connection to the database
$dbh = DBI->connect('DBI:mysql:database='.$LSBDB.';host='.$LSBDBHOST
		.";mysql_opt_read_timeout=432000"
	, $LSBUSER, $LSBDBPASSWD)
    or die "Couldn't connect to database: ".DBI->errstr;

#-----------------------------------------------------------------------

get_Architectures();

LibKB("LibKB");

Distributions();
LibDistrs();

LSB_Libs("x86");
LibAttrs();

#LSB_Ints("x86");
#IntAttrs();

exit;

#-----------------------------------------------------------------------
# Get Architectures list

sub get_Architectures {
	$select = "SELECT Aid, Aname FROM Architecture"
			;

	print "$select\n" if $trace;

	$sth = $dbh->prepare($select) or die "Couldn't prepare $select query: ".DBI->errstr;
	$sth->execute or die "Couldn't execute $select query: ".DBI->errstr;

	while ( my $entry = $sth->fetchrow_hashref ) {
		my ($Aid, $Aname) = @$entry{'Aid', 'Aname'};
		
		$all_archs{$Aname} = $Aid;
	}
	$sth->finish;
}

#-----------------------------------------------------------------------
# Makes list of LSB libraries

sub LSB_Libs {
	my ($arch, $filename) = @_;
	$filename = "LSB_Libs_$arch" if !$filename;
	
	my $arch_request;
	if( $arch =~ /^x86$/i ) {
	    $arch_request = "IA32";
	} else {
		$arch_request = uc $arch;
	}
	
	my $archid = $all_archs{$arch_request}
		or die "Unknown arch: '$arch_request'";
	
	open FILE, ">$filename"
		or die "Can't open '$filename' for writing.";

	print FILE "# Generated file\n";
	print FILE "# List of LSB Libraries\n";
	print FILE "# FORMAT: Lid Lname ALrunname ALappearedin SMmandatorysince ALwithdrawnin Ldeprecatedsince Lcandidatefor\n";
	
	$select = "SELECT DISTINCT Lid, Lname, Lcandidatefor, Ldeprecatedsince,"
					." ALrunname, ALappearedin, ALwithdrawnin, SMmandatorysince FROM Library"
			." LEFT JOIN ArchLib on ALlid=Lid"
			." LEFT JOIN ModLib ON MLlid=Lid "
			." LEFT JOIN SubModule ON SMid=MLmid"
			." WHERE ALaid=$archid OR ALaid=1"
			."	AND ALrunname<>''"
			;

	print "$select\n" if $trace;

	$sth = $dbh->prepare($select) or die "Couldn't prepare $select query: ".DBI->errstr;
	$sth->execute or die "Couldn't execute $select query: ".DBI->errstr;

	while ( my $entry = $sth->fetchrow_hashref ) {
		my ($Lid, $Lname, $Lcandidatefor, $Ldeprecatedsince,
					$ALrunname, $ALappearedin, $ALwithdrawnin, $SMmandatorysince)
			= @$entry{qw/Lid Lname Lcandidatefor Ldeprecatedsince
					ALrunname ALappearedin ALwithdrawnin SMmandatorysince/};
		
		for ($Lid, $Lname, $Lcandidatefor, $Ldeprecatedsince,
				$ALrunname, $ALappearedin, $ALwithdrawnin, $SMmandatorysince)
		{ defined($_) or $_ = ""; }
		
		if( $Lname eq "libstdcxx" ) {
		    $Lname = "libstdc++";
		}

		# TODO: join "\t", ...
		print FILE "$Lid\t$Lname\t$ALrunname\t$ALappearedin\t$SMmandatorysince";
		if ( $ALwithdrawnin or $Ldeprecatedsince or $Lcandidatefor ) {
			print FILE "\t$ALwithdrawnin";
			if ( $Ldeprecatedsince or $Lcandidatefor ) {
				print FILE "\t$Ldeprecatedsince";
				if ( $Lcandidatefor ) {
					print FILE "\t$Lcandidatefor";
				}
			}
		}
		print FILE "\n";
	}
	$sth->finish;
	
	close FILE;
}


#-----------------------------------------------------------------------
# Dump LibraryAttribute table

sub LibAttrs {
	my ($filename) = @_;
	$filename = "LSB_LibAttrs" if !$filename;
	
	open FILE, ">$filename"
		or die "Can't open '$filename' for writing.";

	print FILE "# Generated file\n";
	print FILE "# Dump of the LibraryAttribute table\n";
	print FILE "# FORMAT: LAsoname LAlibname LAtype LAvalue\n";
	
	$select = "SELECT * FROM LibraryAttribute"
			;

	print "$select\n" if $trace;

	$sth = $dbh->prepare($select) or die "Couldn't prepare $select query: ".DBI->errstr;
	$sth->execute or die "Couldn't execute $select query: ".DBI->errstr;

	while ( my $entry = $sth->fetchrow_hashref ) {
		my ($LAsoname, $LAlibname, $LAtype, $LAvalue)
			= @$entry{qw/LAsoname LAlibname LAtype LAvalue/};
		
		for ($LAsoname, $LAlibname, $LAtype, $LAvalue)	{ defined or $_ = ""; }
		
		$LAvalue =~ s/\s+$//; # remove trailing space
		$LAvalue =~ s/(?<!\.)$/./; # Add . at the end
		$LAvalue =~ s/[\n\r]/\x81/g; # Use \x81 instead of \n
		
		print FILE "$LAsoname\t$LAlibname\t$LAtype\t$LAvalue\n";
	}
	$sth->finish;
	
	close FILE;
}

#-----------------------------------------------------------------------
# Makes list of LSB interfaces

sub LSB_Ints {
	my ($arch, $filename) = @_;
	$filename = "LSB_Ints_$arch" if !$filename;
	
	my $arch_request;
	if( $arch =~ /^x86$/i ) {
	    $arch_request = "IA32";
	} else {
		$arch_request = uc $arch;
	}
	
	my $archid = $all_archs{$arch_request}
		or die "Unknown arch: '$arch_request'";
	
	open FILE, ">$filename"
		or die "Can't open '$filename' for writing.";

	print FILE "# Generated file\n";
	print FILE "# List of LSB Interfaces\n";
	print FILE "# FORMAT: Iid Iname Iunmangled Vname Ilibrary AIappearedin SMmandatorysince AIwithdrawnin Ideprecatedsince Icandidatefor\n";
	
	$select = "SELECT DISTINCT Iid, Iname, Iunmangled, Ilibrary, Icandidatefor, Ideprecatedsince,"
					." Vname, AIappearedin, AIwithdrawnin, SMmandatorysince, AIarch FROM Interface"
			." LEFT JOIN LGInt ON LGIint=Iid"
			." LEFT JOIN LibGroup ON LGid=LGIlibg"
			." LEFT JOIN ModLib ON MLlid=LGlib"
			." LEFT JOIN SubModule ON SMid=MLmid"
			." LEFT JOIN ArchInt ON ArchInt.AIint=Iid"
			." LEFT JOIN Version ON Vid=AIversion"
			." WHERE AIarch=$archid OR AIarch=1"
			;

	print "$select\n" if $trace;

	$sth = $dbh->prepare($select) or die "Couldn't prepare $select query: ".DBI->errstr;
	$sth->execute or die "Couldn't execute $select query: ".DBI->errstr;

	while ( my $entry = $sth->fetchrow_hashref ) {
		my ($Iid, $Iname, $Iunmangled, $Vname, $Ilibrary, $AIappearedin, $SMmandatorysince,
				$AIwithdrawnin, $Ideprecatedsince, $Icandidatefor, $AIarch)
			= @$entry{qw/Iid Iname Iunmangled Vname Ilibrary AIappearedin SMmandatorysince
					AIwithdrawnin Ideprecatedsince Icandidatefor AIarch/};
		
		for ($Iid, $Iname, $Iunmangled, $Vname, $Ilibrary, $AIappearedin, $SMmandatorysince,
				$AIwithdrawnin, $Ideprecatedsince, $Icandidatefor, $AIarch)
		{ defined($_) or $_ = ""; }
		
		$Iunmangled = "" if $Iunmangled eq $Iname;
		
		my $generic = ( $AIarch eq "1" ? 1 : 0 );
		print FILE "$Iid\t$Iname\t$Iunmangled\t$Vname\t$generic\t$Ilibrary\t$AIappearedin\t$SMmandatorysince";
		if ( $AIwithdrawnin or $Ideprecatedsince or $Icandidatefor ) {
			print FILE "\t$AIwithdrawnin";
			if ( $Ideprecatedsince or $Icandidatefor ) {
				print FILE "\t$Ideprecatedsince";
				if ( $Icandidatefor ) {
					print FILE "\t$Icandidatefor";
				}
			}
		}
		print FILE "\n";
	}
	$sth->finish;
	
	close FILE;
}


#-----------------------------------------------------------------------
# Get the distribution libraries

my %Libs = ();

sub get_Libs {
	$select = "SELECT DISTINCT RLsoname, Cname, ALsoname, Did FROM RawLibrary"
			." LEFT JOIN Component ON Cid=RLcomponent"
			." LEFT JOIN Distribution ON Did=Cdistr"
			." LEFT JOIN ApprovedLibrary ON ALsoname=RLsoname"
			." WHERE RLsoname!=''"
			." AND Cid IS NOT NULL"
			;

	print "$select\n" if $trace;

	$sth = $dbh->prepare($select) or die "Couldn't prepare $select query: ".DBI->errstr;
	$sth->execute or die "Couldn't execute $select query: ".DBI->errstr;

	while ( my $entry = $sth->fetchrow_hashref ) {
		my $soname = $entry->{'RLsoname'};
		
		defined $entry->{'Did'} or die "Undefined Did: $soname";
		
		$Libs{$soname}{$entry->{'Cname'}}->{APPROVED} = (defined $entry->{'ALsoname'} ? 1 : 0);
		$Libs{$soname}{$entry->{'Cname'}}->{DISTRS}{$entry->{'Did'}} = 1;
	}
	$sth->finish;
}

#-----------------------------------------------------------------------
# Write the LibDistrs file

sub LibDistrs {
	my ($filename) = @_;
	$filename = "LibDistrs" if !$filename;
	
	if ( !%Libs ) { get_Libs(); }
	
	open FILE, ">$filename"
		or die "Can't open '$filename' for writing.";

	print FILE "# Generated file\n";
	print FILE "# List of Distribution Libraries\n";

	# Format: soname \t $cname \t [DistrIDs comma separated] \n
	foreach my $soname ( sort keys %Libs ) {
		foreach my $cname ( sort keys %{$Libs{$soname}} ) {
			my $DistrIDs = "";
			if ( $Libs{$soname}{$cname}->{DISTRS} ) {
				$DistrIDs = join ",", keys %{$Libs{$soname}{$cname}{DISTRS}};
			}
			my $approved = ($Libs{$soname}{$cname}->{APPROVED} or 0);
			
			print FILE join "\t", $soname, $cname, $approved, $DistrIDs;
			print FILE "\n";
		}
	}

	close FILE;
}

#-----------------------------------------------------------------------
# Get the list of distributions

sub Distributions {
	my ($filename) = @_;
	$filename = "Distributions" if !$filename;
	
	$select = "SELECT Did, Dname, Dversion, Aname, Dreleasedate, DVshortname FROM Distribution"
			." LEFT JOIN Architecture ON Aid=Darch"
			." LEFT JOIN DistrVendor ON DVid=Dcompany"
			;

	print "$select\n" if $trace;

	$sth = $dbh->prepare($select) or die "Couldn't prepare $select query: ".DBI->errstr;
	$sth->execute or die "Couldn't execute $select query: ".DBI->errstr;

	open FILE, ">$filename"
		or die "Can't open '$filename' for writing.";

	print FILE "# Generated file\n";
	print FILE "# List of Distributions\n";
	
	# Write distributions to file
	# Format: $Did \t $Dname \t $Dversion \t $arch \t $year \t $vendor \n
	for ( 1 .. $sth->rows ) {
		my $entry = $sth->fetchrow_hashref;
		my $Did = $entry->{'Did'};
		my $Dname = $entry->{'Dname'};
		my $Dversion = $entry->{'Dversion'};
		my $arch = $entry->{'Aname'};
		my $year = substr($entry->{'Dreleasedate'}, 0, 4);
		my $vendor = $entry->{'DVshortname'};

		$arch = 'x86' if $arch eq 'IA32';
		
		print FILE "$Did\t$Dname\t$Dversion\t$arch\t$year\t$vendor\n";
	}
	
	$sth->finish;

	close FILE;
}

sub LibKB {
	my ($filename) = @_;
	$filename = "LibKB" if !$filename;
	
	# Get the distribution libraries
	$select = "SELECT LKBid, LKBlibrary, LKBinfo FROM LibKB"
			;

	print "$select\n" if $trace;

	$sth = $dbh->prepare($select) or die "Couldn't prepare $select query: ".DBI->errstr;
	$sth->execute or die "Couldn't execute $select query: ".DBI->errstr;

	my %LibKB = ();
	my %LKBid_soname = ();

	my $entry;
	while ( $entry = $sth->fetchrow_hashref ) {
		my $id = $entry->{'LKBid'};
		my $info = $entry->{'LKBinfo'};
		
		$LKBid_soname{$id} = $entry->{'LKBlibrary'};
		
		$info =~ s/[\r\n]+$//s;
		
		$LibKB{$id} = $info;
	}
	$sth->finish;

	# Get the distribution libraries
	$select = "SELECT LKBDlkbid, LKBDgroup, LKBDdid FROM LibKBDistr"
			;

	print "$select\n" if $trace;

	$sth = $dbh->prepare($select) or die "Couldn't prepare $select query: ".DBI->errstr;
	$sth->execute or die "Couldn't execute $select query: ".DBI->errstr;

	my %LibKBDistr = ();

	while ( $entry = $sth->fetchrow_hashref ) {
		my $LKBid = $entry->{'LKBDlkbid'};
		my $group = $entry->{'LKBDgroup'};
		my $Did = $entry->{'LKBDdid'};
		
		my $soname = $LKBid_soname{$LKBid};
		
		$LibKBDistr{$LKBid}{$soname}[$group]{$Did} = 1;
	}
	
	$sth->finish;

	open FILE, ">$filename"
		or die "Can't open '$filename' for writing.";

	print FILE "# Generated file\n";
	print FILE "# List of known library portability issues.\n";

	foreach my $LKBid ( keys %LibKBDistr ) {
		print FILE "\n";
		print FILE $LibKB{$LKBid}."\n";
		
		foreach my $soname ( keys %{$LibKBDistr{$LKBid}} ) {
			print FILE "LIB: $soname ";
			my @groups = ();
			foreach my $group ( 1 .. $#{$LibKBDistr{$LKBid}{$soname}} ) {
				if ( $LibKBDistr{$LKBid}{$soname}[$group] ) {
					push @groups, join ",", keys %{$LibKBDistr{$LKBid}{$soname}[$group]};
				} else {
					push @groups, "";
				}
			}
			print FILE join ";", @groups;
			print FILE "\n";
		}
	}

	close FILE;
}

#-----------------------------------------------------------------------
# Interfaces

# Generates file LibIfaces, where Interface IDs grouped by Libs.
# And LibIndex file containing offset of Libs in the LibIfaces.
sub LibIfaces {
	my ($filename, $index_file) = @_;
	$filename = "LibIfaces" if !$filename;
	$index_file = "LibIndex" if !$index_file;
	
	$select = "SELECT RLRIrlid as Lid, RLRIriid as Iid FROM RLibRInt"
			." ORDER BY Lid"
			#." LIMIT 0, 100000" # DBG:
			;

	print "$select\n" if $trace;

	$sth = $dbh->prepare($select) or die "Couldn't prepare $select query: ".DBI->errstr;
	$sth->execute or die "Couldn't execute $select query: ".DBI->errstr;

	open FILE, ">$filename"
		or die "Can't open '$filename' for writing.";

	open INDEX, ">$index_file"
		or die "Can't open '$index_file' for writing.";

	my $cur_lib = 0;
	my $pos = 0;
	while ( my $entry = $sth->fetchrow_hashref ) {
		my ($Lid, $Iid) = @$entry{'Lid', 'Iid'};
		if ( $Lid != $cur_lib ) {
			$cur_lib = $Lid;
			
			print INDEX "$cur_lib|$pos\n";
			
			# DBG:
			my $s = "^$Lid\n";
			$pos += length($s);
			print FILE $s;
			
			print "$Lid\n";
		}
		
		my $s = "$Iid\n";
		$pos += length($s);
		print FILE $s;
	}
	$sth->finish;

	close INDEX;

	close FILE;
}

sub DistrN {
	my ($Did, $filename) = @_;
	$filename = "DistrInts$Did.dat" if !$filename;
	
	# Select Interface IDs for specified DistrID 
	# (joined through the RawLibrary and the Component tables)
	$select = "SELECT DISTINCT RLRIriid as Iid FROM RLibRInt"
			." LEFT JOIN RawLibrary ON RLid=RLRIrlid"
			." LEFT JOIN Component ON Cid=RLcomponent"
			." WHERE Cdistr=$Did"
			." ORDER BY Iid"
			#." LIMIT 0,64" # DBG:
			;

	print "$select\n" if $trace;

	$sth = $dbh->prepare($select) or die "Couldn't prepare $select query: ".DBI->errstr;
	$sth->execute or die "Couldn't execute $select query: ".DBI->errstr;

	open FILE, ">$filename"
		or die "Can't open '$filename' for writing.";

	my $i = 0;
	my $last = -1;
	my $working = 1;
	while ( $working ) {
		# Pack bytes by 8 bits
		my $byte = "";
		for ( my $j = 0; $j < 8; $j++ ) {
			if ( $last < $i ) {
				($last) = $sth->fetchrow_array();
				if ( !defined $last ) { # Out of data
					$working = 0;
					print "Last: $i\n";
					last;
				}
				# DBG:
				#print ">> $last\n";
			}
			if ( $i == $last ) {
				vec( $byte, $j, 1 ) = 1;
				#print ">> $last/$j: ".ord($byte).": ".(join "", unpack("B*", $byte))."\n";
			}
			$i++;
		}
		# Print the byte
		print FILE $byte;
		
		# DBG:
		#print "B(".ord($byte)."): ".(join "", unpack("B*", $byte))."\n";
	}
	
	$sth->finish;

	close FILE;
}

#-----------------------------------------------------------------------
# Interface detailes

sub Interfaces {
	my ($filename, $index_file) = @_;
	$filename = "Interfaces" if !$filename;
	$index_file = "IIndex" if !$index_file;
	
	$select = "SELECT RIid, RIname, RIversion FROM RawInterface"
			." ORDER BY RIname"
			#." LIMIT 0, 1000000" # DBG:
			;

	print "$select\n" if $trace;

	$sth = $dbh->prepare($select) or die "Couldn't prepare $select query: ".DBI->errstr;
	$sth->execute or die "Couldn't execute $select query: ".DBI->errstr;

	open FILE, ">$filename"
		or die "Can't open '$filename' for writing.";

	open INDEX, ">$index_file"
		or die "Can't open '$index_file' for writing.";

	my $last3 = ""; $pos = 0;
	while ( my $entry = $sth->fetchrow_hashref ) {
		my ($RIid, $RIname, $RIversion) = @$entry{'RIid', 'RIname', 'RIversion'}; # NULL: NO
		
		# TODO: RIunmangled
		#my $RIunmangled = ($entry->{'RIunmangled'} or ""); # NULL: YES
		
		$RIname =~ /^.{1,3}/;
		if ( $last3 ne $& ) {
			$last3 = $&;
			print INDEX "$last3|$pos\n";
			
			print "$last3\n";
		}
		
		my $s = "$RIname".($RIversion ? ",".$RIversion : "").";$RIid\n";
		$pos += length($s);
		print FILE $s;
	}
	$sth->finish;

	close INDEX;

	close FILE;
}

#-----------------------------------------------------------------------
# The end

$dbh->disconnect;

print elapsed("ALL:");